# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transformer helper tests."""
from trestle.transforms.transformer_helper import PropertyAccounting
from trestle.transforms.transformer_helper import PropertyManager


def _get_property_set(gid, pid):
    """Generate property test set."""
    return {
        'group': 'group' + str(gid),
        'name': 'name' + str(pid),
        'value': 'value' + str(pid),
        'ns': 'ns' + str(pid),
        'class': 'class' + str(pid),
    }


def _common_test(pa, pm):
    """Reuse common test sequence."""
    ps1 = _get_property_set(1, 1)
    pa.count_group(ps1['group'])
    pa.count_property(ps1['group'], ps1['name'], ps1['value'], ps1['ns'], ps1['class'])
    assert pa.is_common_property(ps1['group'], ps1['name'], ps1['value'], ps1['ns'], ps1['class'])
    ps2 = _get_property_set(1, 2)
    pa.count_group(ps2['group'])
    pa.count_property(ps2['group'], ps2['name'], ps2['value'], ps2['ns'], ps2['class'])
    assert not pa.is_common_property(ps2['group'], ps2['name'], ps2['value'], ps2['ns'], ps2['class'])
    assert not pa.is_common_property(ps1['group'], ps1['name'], ps1['value'], ps1['ns'], ps1['class'])
    pa.count_property(ps1['group'], ps1['name'], ps1['value'], ps1['ns'], ps1['class'])
    pa.count_property(ps2['group'], ps2['name'], ps2['value'], ps2['ns'], ps2['class'])
    assert pa.is_common_property(ps2['group'], ps2['name'], ps2['value'], ps2['ns'], ps2['class'])
    assert pa.is_common_property(ps1['group'], ps1['name'], ps1['value'], ps1['ns'], ps1['class'])


def test_transformer_helper() -> None:
    """Test transformer helper."""
    pa = PropertyAccounting()
    pm = PropertyManager()
    _common_test(pa, pm)


def test_transformer_helper_no_caching() -> None:
    """Test transformer helper with no caching."""
    pa = PropertyAccounting()
    pm = PropertyManager(caching=False)
    _common_test(pa, pm)


def test_transformer_helper_with_checking() -> None:
    """Test transformer with checking."""
    pa = PropertyAccounting()
    pm = PropertyManager(caching=True)
    _common_test(pa, pm)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
#
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#                                                                        #
#        This file is automatically generated by gen_oscal.py            #
#                                                                        #
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#
#
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import AnyUrl, EmailStr, Extra, Field, conint, constr, validator

from trestle.core.base_model import OscalBaseModel
from trestle.oscal import OSCAL_VERSION_REGEX, OSCAL_VERSION
import trestle.oscal.common as common


class Statement(OscalBaseModel):
    """
    Identifies which statements within a control are addressed.
    """

    class Config:
        extra = Extra.forbid

    statement_id: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        alias='statement-id',
        description='A human-oriented identifier reference to a control statement.',
        title='Control Statement Reference',
    )
    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this control statement elsewhere in this or other OSCAL instances. The UUID of the control statement in the source OSCAL instance is sufficient to reference the data item locally or globally (e.g., in an imported OSCAL instance).',
        title='Control Statement Reference Universally Unique Identifier',
    )
    description: str = Field(
        ...,
        description='A summary of how the containing control statement is implemented by the component or capability.',
        title='Statement Implementation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    remarks: Optional[common.Remarks] = None


class State(Enum):
    """
    The operational status.
    """

    under_development = 'under-development'
    operational = 'operational'
    disposition = 'disposition'
    other = 'other'


class SetParameter(OscalBaseModel):
    """
    Identifies the parameter that will be set by the enclosed value.
    """

    class Config:
        extra = Extra.forbid

    param_id: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        alias='param-id',
        description=
        "A human-oriented reference to a parameter within a control, who's catalog has been imported into the current implementation context.",
        title='Parameter ID',
    )
    values: List[common.Value] = Field(...)
    remarks: Optional[common.Remarks] = None


class IncorporatesComponent(OscalBaseModel):
    """
    TBD
    """

    class Config:
        extra = Extra.forbid

    component_uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        alias='component-uuid',
        description='A machine-oriented identifier reference to a component.',
        title='Component Reference',
    )
    description: str = Field(
        ...,
        description='A description of the component, including information about its function.',
        title='Component Description',
    )


class ImportComponentDefinition(OscalBaseModel):
    """
    Loads a component definition from another resource.
    """

    class Config:
        extra = Extra.forbid

    href: str = Field(
        ...,
        description=
        'A link to a resource that defines a set of components and/or capabilities to import into this collection.',
        title='Hyperlink Reference',
    )


class ImplementedRequirement(OscalBaseModel):
    """
    Describes how the containing component or capability implements an individual control.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference a specific control implementation elsewhere in this or other OSCAL instances. The locally defined UUID of the control implementation can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance).This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Control Implementation Identifier',
    )
    control_id: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        alias='control-id',
        description=
        'A human-oriented identifier reference to a control with a corresponding id value. When referencing an externally defined control, the Control Identifier Reference must be used in the context of the external / imported OSCAL instance (e.g., uri-reference).',
        title='Control Identifier Reference',
    )
    description: str = Field(
        ...,
        description=
        'A description of how the specified control is implemented for the containing component or capability.',
        title='Control Implementation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    set_parameters: Optional[List[SetParameter]] = Field(None, alias='set-parameters')
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    statements: Optional[List[Statement]] = Field(None)
    remarks: Optional[common.Remarks] = None


class ControlImplementation(OscalBaseModel):
    """
    Defines how the component or capability supports a set of controls.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference a set of implemented controls elsewhere in this or other OSCAL instances. The locally defined UUID of the control implementation set can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Control Implementation Set Identifier',
    )
    source: str = Field(
        ...,
        description=
        'A reference to an OSCAL catalog or profile providing the referenced control or subcontrol definition.',
        title='Source Resource Reference',
    )
    description: str = Field(
        ...,
        description=
        'A description of how the specified set of controls are implemented for the containing component or capability.',
        title='Control Implementation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    set_parameters: Optional[List[SetParameter]] = Field(None, alias='set-parameters')
    implemented_requirements: List[ImplementedRequirement] = Field(..., alias='implemented-requirements')


class Capability(OscalBaseModel):
    """
    A grouping of other components and/or capabilities.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this capability elsewhere in this or other OSCAL instances. The locally defined UUID of the capability can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance).This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Capability Identifier',
    )
    name: constr(regex=r'^\S(.*\S)?$') = Field(
        ...,
        description="The capability's human-readable name.",
        title='Capability Name',
    )
    description: str = Field(..., description='A summary of the capability.', title='Capability Description')
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    incorporates_components: Optional[List[IncorporatesComponent]] = Field(None, alias='incorporates-components')
    control_implementations: Optional[List[ControlImplementation]] = Field(None, alias='control-implementations')
    remarks: Optional[common.Remarks] = None


class Status(OscalBaseModel):
    """
    Describes the operational status of the system component.
    """

    class Config:
        extra = Extra.forbid

    state: State = Field(..., description='The operational status.', title='State')
    remarks: Optional[common.Remarks] = None


class SystemComponent(OscalBaseModel):
    """
    A defined component that can be part of an implemented system.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this component elsewhere in this or other OSCAL instances. The locally defined UUID of the component can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Component Identifier',
    )
    type: constr(regex=r'^\S(.*\S)?$') = Field(
        ...,
        description='A category describing the purpose of the component.',
        title='Component Type',
    )
    title: str = Field(
        ...,
        description='A human readable name for the system component.',
        title='Component Title',
    )
    description: str = Field(
        ...,
        description='A description of the component, including information about its function.',
        title='Component Description',
    )
    purpose: Optional[str] = Field(
        None,
        description='A summary of the technological or business purpose of the component.',
        title='Purpose',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    status: Status = Field(
        ...,
        description='Describes the operational status of the system component.',
        title='Status',
    )
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    protocols: Optional[List[common.Protocol]] = Field(None)
    remarks: Optional[common.Remarks] = None


class DefinedComponent(OscalBaseModel):
    """
    A defined component that can be part of an implemented system.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this component elsewhere in this or other OSCAL instances. The locally defined UUID of the component can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Component Identifier',
    )
    type: constr(regex=r'^\S(.*\S)?$') = Field(
        ...,
        description='A category describing the purpose of the component.',
        title='Component Type',
    )
    title: str = Field(
        ...,
        description='A human readable name for the component.',
        title='Component Title',
    )
    description: str = Field(
        ...,
        description='A description of the component, including information about its function.',
        title='Component Description',
    )
    purpose: Optional[str] = Field(
        None,
        description='A summary of the technological or business purpose of the component.',
        title='Purpose',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    protocols: Optional[List[common.Protocol]] = Field(None)
    control_implementations: Optional[List[ControlImplementation]] = Field(None, alias='control-implementations')
    remarks: Optional[common.Remarks] = None


class ComponentDefinition(OscalBaseModel):
    """
    A collection of component descriptions, which may optionally be grouped by capability.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this component definition elsewhere in this or other OSCAL instances. The locally defined UUID of the component definition can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Component Definition Universally Unique Identifier',
    )
    metadata: common.Metadata
    import_component_definitions: Optional[List[ImportComponentDefinition]] = Field(
        None, alias='import-component-definitions'
    )
    components: Optional[List[DefinedComponent]] = Field(None)
    capabilities: Optional[List[Capability]] = Field(None)
    back_matter: Optional[common.BackMatter] = Field(None, alias='back-matter')


class Model(OscalBaseModel):
    component_definition: ComponentDefinition = Field(..., alias='component-definition')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle create command."""
import argparse
import json
import pathlib
import sys

from _pytest.monkeypatch import MonkeyPatch

import pytest

from trestle.cli import Trestle
from trestle.common import const
from trestle.common.err import TrestleError
from trestle.core.commands import create
from trestle.oscal.catalog import Catalog

subcommand_list = const.MODEL_TYPE_LIST


@pytest.mark.parametrize('include_optional', [(True), (False)])
def test_create_cmd(tmp_trestle_dir: pathlib.Path, include_optional: bool, monkeypatch: MonkeyPatch) -> None:
    """Happy path test at the cli level."""
    # Test
    testargs_root = ['trestle', 'create']
    for subcommand in subcommand_list:
        name_stem = f'random_named_{subcommand}'
        test_args = testargs_root + ['-t', subcommand] + ['-o', name_stem]
        if include_optional:
            test_args += [const.IOF_SHORT]
        monkeypatch.setattr(sys, 'argv', test_args)
        rc = Trestle().run()
        assert rc == 0
        model_file = tmp_trestle_dir / const.MODEL_TYPE_TO_MODEL_DIR[subcommand] / name_stem / f'{subcommand}.json'
        with open(model_file, 'r', encoding=const.FILE_ENCODING) as fp:
            model = json.load(fp)
        assert name_stem in model[subcommand]['metadata']['title']


def test_no_dir(tmp_empty_cwd: pathlib.Path) -> None:
    """Test for no trestle directory."""
    # Setup argparse
    args = argparse.Namespace(
        trestle_root=tmp_empty_cwd, extension='json', output='catalog', verbose=0, include_optional_fields=False
    )
    with pytest.raises(TrestleError):
        create.CreateCmd.create_object('catalog', Catalog, args)


def test_fail_overwrite(tmp_trestle_dir: pathlib.Path) -> None:
    """Test that a failure occurs when doubling up a create."""
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, extension='json', output='my_catalog', verbose=0, include_optional_fields=False
    )
    rc = create.CreateCmd.create_object('catalog', Catalog, args)
    assert rc == 0
    with pytest.raises(TrestleError):
        create.CreateCmd.create_object('catalog', Catalog, args)


def test_broken_args(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test behaviour on broken arguments."""
    # must be done using sys patching.
    testargs_root = ['trestle', 'create', '-t']
    monkeypatch.setattr(sys, 'argv', testargs_root)
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        Trestle().run()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code > 0
    testargs = testargs_root + ['catalog']
    # missing command
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc > 0
    # missing mandatory args
    testargs = testargs + ['-x', 'json']
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc > 0
    testargs = testargs + ['-o', 'output']
    # correct behavior
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 0
    # correct behavior
    testargs[3] = 'bad_name'
    monkeypatch.setattr(sys, 'argv', testargs)
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        Trestle().run()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code > 0



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ocp4-cis-profile-to-oscal-cd task tests."""

import configparser
import os
import pathlib
import uuid

from _pytest.monkeypatch import MonkeyPatch

from tests.test_utils import text_files_equal

import trestle
import trestle.tasks.ocp4_cis_profile_to_oscal_cd as ocp4_cis_profile_to_oscal_cd
from trestle.tasks.base_task import TaskOutcome


def monkey_uuid_1():
    """Monkey create UUID."""
    return uuid.UUID('56666738-0f9a-4e38-9aac-c0fad00a5821')


def monkey_exception():
    """Monkey exception."""
    raise Exception('foobar')


def monkey_trestle_version():
    """Monkey trestle version."""
    return '0.21.0'


def test_ocp4_cis_profile_to_oscal_cd_print_info(tmp_path: pathlib.Path):
    """Test print_info call."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.print_info()
    assert retval is None


def test_ocp4_cis_profile_to_oscal_cd_simulate(tmp_path: pathlib.Path):
    """Test simulate call."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.simulate()
    assert retval == TaskOutcome.SIM_SUCCESS
    assert len(os.listdir(str(tmp_path))) == 0


def test_ocp4_cis_profile_to_oscal_cd_execute(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test execute call."""
    monkeypatch.setattr(uuid, 'uuid4', monkey_uuid_1)
    monkeypatch.setattr(trestle, '__version__', monkey_trestle_version())
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    d_expected = pathlib.Path(section['output-dir'])
    d_produced = tmp_path
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    tgt.set_timestamp('2021-07-19T14:03:03.000+00:00')
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    list_dir = os.listdir(d_produced)
    assert len(list_dir) == 1
    assert d_expected != d_produced
    for fn in list_dir:
        f_expected = d_expected / fn
        f_produced = d_produced / fn
        result = text_files_equal(f_expected, f_produced)
        assert (result)


def test_ocp4_cis_profile_to_oscal_cd_execute_selected_rules2(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test execute selected rules call."""
    monkeypatch.setattr(uuid, 'uuid4', monkey_uuid_1)
    monkeypatch.setattr(trestle, '__version__', monkey_trestle_version())
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    d_expected = pathlib.Path(section['output-dir'])
    d_produced = tmp_path
    section['output-dir'] = str(tmp_path)
    section['selected-rules'] = 'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/extras/selected_rules2.json'
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    tgt.set_timestamp('2021-07-19T14:03:03.000+00:00')
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    list_dir = os.listdir(d_produced)
    assert len(list_dir) == 1
    assert d_expected != d_produced
    for fn in list_dir:
        f_expected = d_expected / fn
        f_produced = d_produced / fn
        result = text_files_equal(f_expected, f_produced)
        assert (result)


def test_ocp4_cis_profile_to_oscal_cd_execute_enabled_rules2(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test execute enabled rules call."""
    monkeypatch.setattr(uuid, 'uuid4', monkey_uuid_1)
    monkeypatch.setattr(trestle, '__version__', monkey_trestle_version())
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    d_expected = pathlib.Path(section['output-dir'])
    d_produced = tmp_path
    section['output-dir'] = str(tmp_path)
    section['enabled-rules'] = 'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/extras/enabled_rules2.json'
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    tgt.set_timestamp('2021-07-19T14:03:03.000+00:00')
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    list_dir = os.listdir(d_produced)
    assert len(list_dir) == 1
    assert d_expected != d_produced
    for fn in list_dir:
        f_expected = d_expected / fn
        f_produced = d_produced / fn
        result = text_files_equal(f_expected, f_produced)
        assert (result)


def test_ocp4_cis_profile_to_oscal_cd_execute_enabled_rules3(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test execute enabled rules call."""
    monkeypatch.setattr(uuid, 'uuid4', monkey_uuid_1)
    monkeypatch.setattr(trestle, '__version__', monkey_trestle_version())
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    d_expected = pathlib.Path(section['output-dir'])
    d_produced = tmp_path
    section['output-dir'] = str(tmp_path)
    section['enabled-rules'] = 'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/extras/enabled_rules3.json'
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    tgt.set_timestamp('2021-07-19T14:03:03.000+00:00')
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    list_dir = os.listdir(d_produced)
    assert len(list_dir) == 1
    assert d_expected != d_produced


def test_ocp4_cis_profile_to_oscal_cd_bogus_config(tmp_path: pathlib.Path):
    """Test execute call bogus config."""
    section = None
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_list(tmp_path: pathlib.Path):
    """Test execute call missing profile-list."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section.pop('profile-list')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_component_name(tmp_path: pathlib.Path):
    """Test execute call missing component-name."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section.pop('component-name')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_type(tmp_path: pathlib.Path):
    """Test execute call missing profile-type."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section.pop('profile-type')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_ns(tmp_path: pathlib.Path):
    """Test execute call missing profile-ns."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section.pop('profile-ns')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_key(tmp_path: pathlib.Path):
    """Test execute missing profile-file."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    profile_list = section['profile-list'].split()
    for profile in profile_list:
        section.pop(f'profile-file.{profile}')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_file(tmp_path: pathlib.Path):
    """Test execute missing profile-file."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    profile_list = section['profile-list'].split()
    for profile in profile_list:
        section[f'profile-file.{profile}'] = '/foobar'
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_url(tmp_path: pathlib.Path):
    """Test execute missinf profile-url."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    profile_list = section['profile-list'].split()
    for profile in profile_list:
        section.pop(f'profile-url.{profile}')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_profile_title(tmp_path: pathlib.Path):
    """Test execute call missing profile-title."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    profile_list = section['profile-list'].split()
    for profile in profile_list:
        section.pop(f'profile-title.{profile}')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_output_dir(tmp_path: pathlib.Path):
    """Test execute call missing output-dir."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section.pop('output-dir')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_no_overwrite(tmp_path: pathlib.Path):
    """Test execute no overwrite."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path('tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd.config')
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    section['output-overwrite'] = 'false'
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_duplicate_rule(tmp_path: pathlib.Path):
    """Test execute duplicate rule exists."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd2.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS
    section['output-overwrite'] = 'false'
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_exception(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test _get_cis_rules exception."""
    monkeypatch.setattr(ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD, '_get_cis_rules', monkey_exception)
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd2.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_rules_section(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test missing section selected-rules."""
    monkeypatch.setattr(ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD, '_get_cis_rules', monkey_exception)
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd2.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section.pop('selected-rules')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_rules_file(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test missing file enabled-rules."""
    monkeypatch.setattr(ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD, '_get_cis_rules', monkey_exception)
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd2.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section['enabled-rules'] = '/foobar'
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_parameters_key(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test missing file enabled-rules."""
    monkeypatch.setattr(ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD, '_get_cis_rules', monkey_exception)
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd2.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section.pop('rule-to-parameters-map')
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_missing_parameters_file(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test missing file enabled-rules."""
    monkeypatch.setattr(ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD, '_get_cis_rules', monkey_exception)
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd2.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    section['rule-to-parameters-map'] = '/foobar'
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_ocp4_cis_profile_to_oscal_cd_bogus_rules_file(tmp_path: pathlib.Path):
    """Test bogus rules."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(
        'tests/data/tasks/ocp4-cis-profile-to-oscal-cd/test-ocp4-cis-profile-to-oscal-cd3.config'
    )
    config.read(config_path)
    section = config['task.ocp4-cis-profile-to-oscal-cd']
    section['output-dir'] = str(tmp_path)
    tgt = ocp4_cis_profile_to_oscal_cd.Ocp4CisProfileToOscalCD(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.SUCCESS



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create resolved catalog from profile."""

import logging
import re
from typing import Dict, Iterator, List, Optional, Set
from uuid import uuid4

import trestle.oscal.catalog as cat
import trestle.oscal.profile as prof
from trestle.common.const import MARKDOWN_URL_REGEX, UUID_REGEX
from trestle.common.err import TrestleError
from trestle.common.list_utils import none_if_empty
from trestle.core.catalog_interface import CatalogInterface
from trestle.core.pipeline import Pipeline
from trestle.oscal import common

logger = logging.getLogger(__name__)


class Prune(Pipeline.Filter):
    """Prune the catalog based on the import include rule."""

    def __init__(self, import_: prof.Import, profile: prof.Profile) -> None:
        """
        Inject the import.

        This needs to be created prior to knowing the catalog.
        The profile itself is only needed for debug messages.
        The import is one possibly several imports in that profile.
        """
        self._import = import_
        self._profile = profile
        self._catalog_interface: Optional[CatalogInterface] = None
        self._catalog: Optional[cat.Catalog] = None

    def _set_catalog(self, catalog: cat.Catalog) -> None:
        """Set the catalog used by the catalog interface."""
        self._catalog_interface = CatalogInterface(catalog)
        self._catalog = catalog

    def _find_uuid_refs(self, control_id: str) -> Set[str]:
        """
        Find all needed resource refs buried in control links and prose.

        For any controls retained in the resolved profile catalog, if any
        prose references a document by uuid, that reference needs to be in backmatter.
        """
        control = self._catalog_interface.get_control(control_id)
        refs = set()
        if control.links is not None:
            for link in control.links:
                uuid_str = link.href.replace('#', '')
                refs.add(uuid_str)
        if control.parts is not None:
            for part in control.parts:
                if part.prose is not None:
                    # find the two parts, label and ref, in each markdown url
                    # expecting form [label](#uuid)
                    # but if it is a control ref it may be e.g. [CM-7](#cm-7)
                    # for now label is not used
                    # the ref may be a uuid or control id
                    # currently only uuids are used to confirm needed presence in backmatter
                    # note that prose may be multi-line but findall searches all lines
                    matches = re.findall(MARKDOWN_URL_REGEX, part.prose)
                    for match in matches:
                        ref = match[1]
                        if len(ref) > 1 and ref[0] == '#':
                            uuid_match = re.findall(UUID_REGEX, ref[1:])
                            # there should be only one uuid in the parens
                            if uuid_match:
                                refs.add(uuid_match[0])
        if control.controls is not None:
            for sub_control in control.controls:
                refs.update(self._find_uuid_refs(sub_control.id))
        return refs

    def _find_all_uuid_refs(self, needed_control_ids: List[str]) -> Set[str]:
        """Find all references needed by controls."""
        refs = set()
        for control_id in needed_control_ids:
            refs.update(self._find_uuid_refs(control_id))
        return refs

    def _controls_selected(self, select_list: Optional[List[prof.SelectControlById]]) -> List[str]:
        control_ids: List[str] = []
        if select_list is not None:
            for select_control in select_list:
                if select_control.matching is not None:
                    raise TrestleError('Profiles with SelectControlById based on matching are not supported.')
                include_children = select_control.with_child_controls == prof.WithChildControls.yes
                if select_control.with_ids:
                    new_ids = [withid.__root__ for withid in select_control.with_ids]
                    for id_ in new_ids:
                        control_ids.append(id_)
                        if include_children:
                            control_ids.extend(self._catalog_interface.get_dependent_control_ids(id_))
        return control_ids

    def _find_needed_control_ids(self) -> List[str]:
        """Get list of control_ids needed by profile and corresponding groups."""
        if self._import.include_controls is not None:
            include_ids = self._controls_selected(self._import.include_controls)
        else:
            if self._import.include_all is None:
                logger.warning('Profile does not specify include-controls, so including all.')
            include_ids = self._catalog_interface.get_control_ids()

        exclude_ids = self._controls_selected(self._import.exclude_controls)

        if not set(include_ids).issuperset(set(exclude_ids)):
            logger.debug(f'include_ids is not a superset of exclude_ids in import {self._import.href}')
        return [id_ for id_ in include_ids if id_ not in exclude_ids]

    def _prune_control(self, needed_ids: List[str], control: cat.Control, exclude_ids: List[str]) -> cat.Control:
        """
        Prune the control based on the Import requirements.

        This is only called if the control is needed
        Some or all of its sub_controls may not be needed
        This always returns the original control, possibly with fewer subcontrols
        """
        if control.controls is None:
            return control
        controls = []
        for sub_control in control.controls:
            if sub_control.id in needed_ids and sub_control.id not in exclude_ids:
                controls.append(self._prune_control(needed_ids, sub_control, exclude_ids))
                exclude_ids.append(sub_control.id)
        control.controls = none_if_empty(controls)
        return control

    def _prune_controls(self, needed_ids: List[str]) -> List[str]:
        loaded_ids = []
        final_ids: List[str] = []
        for control_id in needed_ids:
            if control_id not in loaded_ids:
                control = self._catalog_interface.get_control(control_id)
                if control is None:
                    msg = (
                        f'Profile titled "{self._profile.metadata.title}" references control {control_id} '
                        f'but it is not in catalog titled "{self._catalog.metadata.title}"'
                    )
                    raise TrestleError(msg)
                control = self._prune_control(needed_ids, control, loaded_ids)
                self._catalog_interface.replace_control(control)
                loaded_ids.append(control_id)
                final_ids.append(control_id)
        return final_ids

    def _prune_catalog(self) -> cat.Catalog:
        """Prune the controls in the current catalog."""
        if self._import is None:
            return self._catalog

        needed_ids = self._find_needed_control_ids()

        # if a control includes controls - only include those that we know are needed
        final_control_ids = self._prune_controls(needed_ids)

        cat_controls = []

        # build the needed groups of controls
        group_dict: Dict[str, cat.Group] = {}
        for control_id in final_control_ids:
            control = self._catalog_interface.get_control(control_id)
            group_id, group_title, group_class = self._catalog_interface.get_group_info_by_control(control_id)
            if not group_id:
                cat_controls.append(control)
                continue
            group = group_dict.get(group_id)
            if group is None:
                group = cat.Group(id=group_id, title=group_title, class_=group_class, controls=[control])
                group_dict[group_id] = group
            else:
                group_dict[group_id].controls.append(control)

        # find all referenced uuids - they should be 1:1 with those in backmatter
        needed_uuid_refs: Set[str] = self._find_all_uuid_refs(final_control_ids)

        # prune the list of resources to only those that are needed
        new_resources: Optional[List[common.Resource]] = []
        if self._catalog.back_matter is not None and self._catalog.back_matter.resources is not None:
            new_resources = [res for res in self._catalog.back_matter.resources if res.uuid in needed_uuid_refs]

        new_groups: Optional[List[cat.Group]] = list(group_dict.values())

        # should avoid empty lists so set to None if empty
        new_resources = none_if_empty(new_resources)
        new_groups = none_if_empty(new_groups)
        cat_controls = none_if_empty(cat_controls)
        new_params = self._catalog.params

        new_cat = cat.Catalog(
            uuid=str(uuid4()),
            metadata=self._catalog.metadata,
            back_matter=common.BackMatter(resources=new_resources),
            controls=cat_controls,
            groups=new_groups,
            params=new_params
        )

        return new_cat

    def process(self, catalog_iter: Iterator[cat.Catalog]) -> Iterator[cat.Catalog]:
        """
        Prune the catalog based on the include rule in the import_.

        This only processes the one catalog yielded by the one import in this pipeline.
        It must yield in order to have the merge filter loop over available imported catalogs.
        """
        self._set_catalog(next(catalog_iter))
        logger.debug(f'prune yielding catalog {self._catalog.metadata.title} with import {self._import.href}')
        yield self._prune_catalog()



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create resolved catalog from profile."""

import logging
import re
from typing import Dict, Iterator, List, Optional

import trestle.oscal.catalog as cat
import trestle.oscal.profile as prof
from trestle.common.common_types import OBT
from trestle.common.err import TrestleError, TrestleNotFoundError
from trestle.common.list_utils import as_list
from trestle.core.catalog_interface import CatalogInterface
from trestle.core.control_io import ControlIOReader, ParameterRep
from trestle.core.pipeline import Pipeline
from trestle.oscal import common

logger = logging.getLogger(__name__)


class Modify(Pipeline.Filter):
    """Modify the controls based on the profile."""

    def __init__(
        self,
        profile: prof.Profile,
        change_prose: bool = False,
        block_adds: bool = False,
        block_params: bool = False,
        params_format: str = None,
        param_rep: ParameterRep = ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) -> None:
        """Initialize the filter."""
        self._profile = profile
        self._catalog_interface: Optional[CatalogInterface] = None
        self._block_adds = block_adds
        self._block_params = block_params
        self._change_prose = change_prose
        self._params_format = params_format
        self._param_rep = param_rep
        logger.debug(f'modify initialize filter with profile {profile.metadata.title}')

    @staticmethod
    def _replace_ids_with_text(prose: str, param_rep: ParameterRep, param_dict: Dict[str, common.Parameter]) -> str:
        """Find all instances of param_ids in prose and replace each with corresponding parameter representation.

        Need to check all values in dict for a match
        Reject matches where the string has an adjacent alphanumeric char: param_1 and param_10 or aparam_1
        """
        for param in param_dict.values():
            if param.id not in prose:
                continue
            # create the replacement text for the param_id
            param_str = ControlIOReader.param_to_str(param, param_rep)
            # non-capturing groups are odd in re.sub so capture all 3 groups and replace the middle one
            pattern = r'(^|[^a-zA-Z0-9_])' + param.id + r'($|[^a-zA-Z0-9_])'
            prose = re.sub(pattern, r'\1' + param_str + r'\2', prose)
        return prose

    @staticmethod
    def _replace_params(
        text: str,
        param_dict: Dict[str, common.Parameter],
        params_format: Optional[str] = None,
        param_rep: ParameterRep = ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) -> str:
        """
        Replace params found in moustaches with values from the param_dict.

        A single line of prose may contain multiple moustaches.
        """
        # first check if there are any moustache patterns in the text
        if param_rep == ParameterRep.LEAVE_MOUSTACHE:
            return text
        staches: List[str] = re.findall(r'{{.*?}}', text)
        if not staches:
            return text
        # now have list of all staches including braces, e.g. ['{{foo}}', '{{bar}}']
        # clean the staches so they just have the param ids
        param_ids = []
        for stache in staches:
            # remove braces so these are just param_ids but may have extra chars
            stache_contents = stache[2:(-2)]
            param_id = stache_contents.replace('insert: param,', '').strip()
            param_ids.append(param_id)

        # now replace original stache text with param values
        for i, _ in enumerate(staches):
            # A moustache may refer to a param_id not listed in the control's params
            if param_ids[i] not in param_dict:
                logger.warning(f'Control prose references param {param_ids[i]} not set in the control: {staches}')
            elif param_dict[param_ids[i]] is not None:
                param = param_dict[param_ids[i]]
                param_str = ControlIOReader.param_to_str(param, param_rep, False, False, params_format)
                text = text.replace(staches[i], param_str, 1).strip()
                if param_rep != ParameterRep.LABEL_OR_CHOICES and not param.values:
                    logger.warning(f'Parameter {param_id} has no values and was referenced by prose.')
            else:
                logger.warning(f'Control prose references param {param_ids[i]} with no specified value.')
        return text

    @staticmethod
    def _replace_part_prose(
        control: cat.Control,
        part: common.Part,
        param_dict: Dict[str, common.Parameter],
        params_format: Optional[str] = None,
        param_rep: ParameterRep = ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) -> None:
        """Replace the part prose according to set_param."""
        if part.prose is not None:
            fixed_prose = Modify._replace_params(part.prose, param_dict, params_format, param_rep)
            # change the prose in the control itself
            part.prose = fixed_prose
        for prt in as_list(part.parts):
            Modify._replace_part_prose(control, prt, param_dict, params_format, param_rep)
        for sub_control in as_list(control.controls):
            for prt in as_list(sub_control.parts):
                Modify._replace_part_prose(sub_control, prt, param_dict, params_format, param_rep)

    @staticmethod
    def _replace_control_prose(
        control: cat.Control,
        param_dict: Dict[str, common.Parameter],
        params_format: Optional[str] = None,
        param_rep: ParameterRep = ParameterRep.VALUE_OR_LABEL_OR_CHOICES
    ) -> None:
        """Replace the control prose according to set_param."""
        for param in as_list(control.params):
            Modify._replace_param_choices(param, param_dict)
        for part in as_list(control.parts):
            if part.prose is not None:
                fixed_prose = Modify._replace_params(part.prose, param_dict, params_format, param_rep)
                # change the prose in the control itself
                part.prose = fixed_prose
            for prt in as_list(part.parts):
                Modify._replace_part_prose(control, prt, param_dict, params_format, param_rep)

    @staticmethod
    def _add_contents_as_list(add: prof.Add) -> List[OBT]:
        add_list = []
        add_list.extend(as_list(add.props))
        add_list.extend(as_list(add.parts))
        add_list.extend(as_list(add.links))
        return add_list

    @staticmethod
    def _add_adds_to_part(part: common.Part, add: prof.Add) -> None:
        for attr in ['params', 'props', 'parts', 'links']:
            add_list = getattr(add, attr, None)
            if add_list:
                Modify._add_attr_to_part(part, add_list, attr, add.position)

    @staticmethod
    def _add_to_list(input_list: List[OBT], add: prof.Add) -> bool:
        """Add the contents of the add according to its by_id and position.

        Return True on success or False if id needed and not found.

        This is only called when by_id is not None.
        The add will be inserted if the id is found, or return False if not.
        This allows a separate recursive routine to search sub-lists for the id.
        """
        add_list = Modify._add_contents_as_list(add)
        # Test here for matched by_id attribute.
        try:
            for index in range(len(input_list)):
                if input_list[index].id == add.by_id:
                    if add.position == prof.Position.after:
                        for offset, new_item in enumerate(add_list):
                            input_list.insert(index + 1 + offset, new_item)
                        return True
                    elif add.position == prof.Position.before:
                        for offset, new_item in enumerate(add_list):
                            input_list.insert(index + offset, new_item)
                        return True
                    # if starting or ending, the adds go directly into this part according to type
                    Modify._add_adds_to_part(input_list[index], add)
                    return True
        except AttributeError:
            raise TrestleError(
                'Cannot use "after" or "before" modifications for a list where elements'
                + ' do not contain the referenced by_id attribute.'
            )
        return False

    @staticmethod
    def _add_to_parts(parts: List[common.Part], add: prof.Add) -> bool:
        """
        Add the add to the parts.

        This is only called if add.by_id is not None.
        """
        if Modify._add_to_list(parts, add):
            return True
        for part in parts:
            if part.parts is not None and Modify._add_to_parts(part.parts, add):
                return True
        return False

    @staticmethod
    def _add_attr_to_part(part: common.Part, items: List[OBT], attr: str, position: prof.Position) -> None:
        attr_list = as_list(getattr(part, attr, None))
        if position in [prof.Position.starting, prof.Position.before]:
            items.extend(attr_list)
            attr_list = items
        else:
            attr_list.extend(items)
        setattr(part, attr, attr_list)

    @staticmethod
    def _add_attr_to_control(control: cat.Control, items: List[OBT], attr: str, position: prof.Position) -> None:
        attr_list = as_list(getattr(control, attr, None))
        if position in [prof.Position.starting, prof.Position.before]:
            items.extend(attr_list)
            attr_list = items
        else:
            attr_list.extend(items)
        setattr(control, attr, attr_list)

    @staticmethod
    def _add_to_control(control: cat.Control, add: prof.Add) -> None:
        """First step in applying Add to control."""
        control.parts = as_list(control.parts)
        if add.by_id is None or add.by_id == control.id:
            # add contents will be added to the control directly and with no recursion
            for attr in ['params', 'props', 'parts', 'links']:
                add_list = getattr(add, attr, None)
                if add_list:
                    Modify._add_attr_to_control(control, add_list, attr, add.position)
            return
        else:
            # this is only called if by_id is not None
            if not Modify._add_to_parts(control.parts, add):
                logger.warning(f'Could not find id for add in control {control.id}: {add.by_id}')

    @staticmethod
    def _set_overwrite_items(param: common.Parameter, set_param: prof.SetParameter) -> None:
        # these overwrite
        if set_param.class_:
            param.class_ = set_param.class_
        if set_param.depends_on:
            param.depends_on = set_param.depends_on
        if set_param.label:
            param.label = set_param.label
        if set_param.usage:
            param.usage = set_param.usage
        if set_param.values:
            param.values = set_param.values
        if set_param.select:
            param.select = set_param.select

    @staticmethod
    def _set_appended_items(param: common.Parameter, set_param: prof.SetParameter) -> None:
        # these append
        if set_param.constraints:
            if not param.constraints:
                param.constraints = []
            param.constraints.extend(set_param.constraints)
        if set_param.guidelines:
            if not param.guidelines:
                param.guidelines = []
            param.guidelines.extend(set_param.guidelines)

    @staticmethod
    def _set_replaced_or_appended_items(param: common.Parameter, set_param: prof.SetParameter) -> None:
        # these replace or append
        if set_param.props:
            new_props = as_list(param.props)
            names = [prop.name for prop in new_props]
            for prop in set_param.props:
                if prop.name in names:
                    new_props[names.index(prop.name)] = prop
                else:
                    new_props.append(prop)
            param.props = new_props
        if set_param.links:
            new_links = as_list(param.links)
            hrefs = [link.href for link in new_links]
            for link in set_param.links:
                if link.href in hrefs:
                    new_links[hrefs.index(link.href)] = link
                else:
                    new_links.append(link)
            param.links = new_links

    def _set_parameter_in_control_or_loose(self, set_param: prof.SetParameter) -> None:
        """
        Find the control with the param_id in it and set the parameter contents.

        It modifies controls in the control_dict not the catalog.
        Parameters are either bound to a control or are 'loose' and bound to the catalog itself.
        """
        # find the target param in control or the catalog's loose ones, i.e. catalog.params
        control = self._catalog_interface.get_control_by_param_id(set_param.param_id)
        loose_param = False
        if control:
            control.params = as_list(control.params)
            param_ids = [param.id for param in control.params]
            if set_param.param_id not in param_ids:
                raise TrestleNotFoundError(f'Param id {set_param.param_id} not found in control {control.id}')
            index = param_ids.index(set_param.param_id)
            param = control.params[index]
        else:
            param = self._catalog_interface.loose_param_dict.get(set_param.param_id, None)
            if param:
                loose_param = True
            else:
                logger.warning(f'SetParameter for param_id {set_param.param_id} not found in catalog')
                return

        # rules here follow https://pages.nist.gov/OSCAL/concepts/processing/profile-resolution/
        # see 'Modify Phase' and Setting Parameters

        Modify._set_overwrite_items(param, set_param)
        Modify._set_appended_items(param, set_param)
        Modify._set_replaced_or_appended_items(param, set_param)

        if loose_param:
            self._catalog_interface.loose_param_dict[set_param.param_id] = param
        else:
            control.params[index] = param
            self._catalog_interface.replace_control(control)

    def _change_prose_with_param_values(self):
        """Go through all controls and change prose based on param values."""
        param_dict: Dict[str, common.Parameter] = {}
        # build the full mapping of params to values from the catalog interface
        for control in self._catalog_interface.get_all_controls_from_dict():
            param_dict.update(ControlIOReader.get_control_param_dict(control, False))
        param_dict.update(self._catalog_interface.loose_param_dict)
        # insert param values into prose of all controls
        for control in self._catalog_interface.get_all_controls_from_dict():
            self._replace_control_prose(control, param_dict, self._params_format, self._param_rep)

    @staticmethod
    def _replace_param_choices(param: common.Parameter, param_dict: Dict[str, common.Parameter]) -> None:
        """Set values for all choices param that refer to params with values."""
        if param.select:
            new_choices: List[str] = []
            for choice in as_list(param.select.choice):
                new_choice = Modify._replace_params(choice, param_dict)
                new_choices.append(new_choice)
            param.select.choice = new_choices

    def _modify_controls(self, catalog: cat.Catalog) -> cat.Catalog:
        """Modify the controls based on the profile."""
        logger.debug(f'modify specify catalog {catalog.metadata.title} for profile {self._profile.metadata.title}')
        self._catalog_interface = CatalogInterface(catalog)
        alters: Optional[List[prof.Alter]] = None
        # find the modify and alters
        if self._profile.modify:
            # change all parameter values
            if self._profile.modify.set_parameters and not self._block_params:
                set_param_list = self._profile.modify.set_parameters
                for set_param in set_param_list:
                    self._set_parameter_in_control_or_loose(set_param)
            alters = self._profile.modify.alters

        if alters is not None:
            title = self._profile.metadata.title
            for alter in alters:
                if alter.control_id is None:
                    logger.warning(f'Alter must have control id specified in profile {title}.')
                    continue
                id_ = alter.control_id
                if alter.removes is not None:
                    logger.warning(f'Alter not supported for removes in profile {title} control {id_}')
                    continue
                # we want a warning about adds even if adds are blocked, as in profile generate
                if alter.adds is None:
                    logger.warning(f'Alter has no adds in profile {title} control {id_}')
                    continue
                if not self._block_adds:
                    for add in alter.adds:
                        if add.position is None and add.parts is not None:
                            msg = f'Alter/Add position is not specified in profile {title} control {id_}'
                            msg += ' when adding part, so defaulting to ending.'
                            logger.warning(msg)
                            add.position = prof.Position.ending
                        control = self._catalog_interface.get_control(id_)
                        if control is None:
                            logger.warning(
                                f'Alter/Add refers to control {id_} but it is not found in the import '
                                + f'for profile {self._profile.metadata.title}'
                            )
                        else:
                            self._add_to_control(control, add)
                            self._catalog_interface.replace_control(control)

        if self._change_prose:
            # go through all controls and fix the prose based on param values
            self._change_prose_with_param_values()

        catalog = self._catalog_interface.get_catalog()

        # update the original profile metadata with new contents
        # roles and responsible-parties will be pulled in with new uuid's
        new_metadata = self._profile.metadata
        new_metadata.title = f'{catalog.metadata.title}: Resolved by profile {self._profile.metadata.title}'
        links: List[common.Link] = []
        for import_ in self._profile.imports:
            links.append(common.Link(**{'href': import_.href, 'rel': 'resolution-source'}))
        new_metadata.links = links

        # move catalog controls from dummy group '' into the catalog
        for group in as_list(catalog.groups):
            if not group.id:
                catalog.controls = group.controls
                catalog.groups.remove(group)
                break

        catalog.metadata = new_metadata

        return catalog

    def process(self, catalog_iter: Iterator[cat.Catalog]) -> Iterator[cat.Catalog]:
        """Make the modifications to the controls based on the profile."""
        catalog = next(catalog_iter)
        logger.debug(
            f'modify process with catalog {catalog.metadata.title} using profile {self._profile.metadata.title}'
        )
        yield self._modify_controls(catalog)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle author headers command."""

import pathlib
import shutil
import sys
from uuid import uuid4

from _pytest.monkeypatch import MonkeyPatch

import pytest

from trestle.cli import Trestle
from trestle.common import file_utils
from trestle.core.commands.author.consts import START_TEMPLATE_VERSION


@pytest.mark.parametrize(
    'task_name, template_path, content_path, global_, setup_rc, template_validate_rc, validate_rc',
    [
        ('catalogs', pathlib.Path('irrelevant'), pathlib.Path('irrelevant_2'), False, 1, 1, 1),
        (
            'test_tasks_1',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/meets_templates'),
            False,
            0,
            0,
            0
        ),
        (
            'test_tasks_2',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            0,
            0,
            0
        ),
        (
            'test_tasks_3',
            pathlib.Path('author/headers/bad_templates_drawio'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            0,
            1,
            1
        ),
        (
            'test_tasks_4',
            pathlib.Path('author/headers/bad_templates_wrong_names'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            0,
            1,
            1
        ),
        (
            'test_tasks_5',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_drawio'),
            False,
            0,
            0,
            3
        ),
        (
            'test_tasks_6',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_drawio_bad_file'),
            False,
            0,
            0,
            1
        ),
        (
            'test_tasks_7',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_md'),
            False,
            0,
            0,
            3
        ),
        (
            'test_tasks_8',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_md_bad_file'),
            False,
            0,
            0,
            3
        ),
        (
            'test_tasks_9',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/passes_with_extraneous_files'),
            False,
            0,
            0,
            0
        ),
        (
            'test_tasks_10',
            pathlib.Path('author/headers/bad_templates_md'),
            pathlib.Path('author/headers/passes_with_extraneous_files'),
            False,
            0,
            1,
            1
        ),
        (
            'test_tasks_11',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            True,
            0,
            0,
            0
        ),
        (
            None,
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            True,
            0,
            0,
            0
        ),
        (
            None,
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            1,
            1,
            1
        ),
    ]
)
def test_e2e(
    task_name: str,
    template_path: pathlib.Path,
    content_path: pathlib.Path,
    global_: bool,
    setup_rc: int,
    template_validate_rc: int,
    validate_rc: int,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Run an E2E workflow with a number of test criteria."""
    command_string_setup = 'trestle author headers setup'
    command_string_create_sample = 'trestle author headers create-sample'
    command_string_validate_template = 'trestle author headers template-validate'
    command_string_validate_content = 'trestle author headers validate -r'
    template_target_loc: pathlib.Path
    if task_name:
        tn_string = f' -tn {task_name}'
        command_string_setup += tn_string
        command_string_create_sample += tn_string
        command_string_validate_template += tn_string
        command_string_validate_content += tn_string
        template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / START_TEMPLATE_VERSION
    else:
        task_name = 'placeholder'
    if global_:
        global_str = ' --global'
        command_string_setup += global_str
        command_string_create_sample += global_str
        command_string_validate_template += global_str
        command_string_validate_content += global_str
        template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / '__global__' / START_TEMPLATE_VERSION
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}'
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = Trestle().run()
    assert rc == setup_rc
    if setup_rc > 0:
        return
    shutil.rmtree(str(template_target_loc))
    shutil.copytree(str(testdata_dir / template_path), str(template_target_loc))

    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    rc = Trestle().run()
    assert rc == template_validate_rc
    if template_validate_rc > 0:
        return

    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    rc = Trestle().run()
    assert rc == 0

    shutil.copytree(str(testdata_dir / content_path), str(test_content_loc))
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == validate_rc


def test_abort_safely_on_missing_directory(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that validation fails cleanly on a missing directory."""
    task_name = 'tester'
    command_string_setup = f'trestle author headers setup -tn {task_name}'
    command_string_validate_content = f'trestle author headers validate -tn {task_name} -r'
    task_dir = tmp_trestle_dir / task_name

    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = Trestle().run()
    assert rc == 0
    shutil.rmtree(str(task_dir))

    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc > 0


def test_taskpath_is_a_file(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that validation fails cleanly on a missing directory."""
    task_name = 'tester'
    command_string_setup = f'trestle author headers setup -tn {task_name}'
    command_string_validate_content = f'trestle author headers validate -tn {task_name} -r'
    task_dir = tmp_trestle_dir / task_name

    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = Trestle().run()
    assert rc == 0

    shutil.rmtree(str(task_dir))

    task_dir.touch()
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc > 0


def test_taskpath_is_a_file_setup(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that setup fails cleanly when taskpath is a file."""
    task_name = 'tester'
    command_string_setup = f'trestle author headers setup -tn {task_name}'
    task_dir = tmp_trestle_dir / task_name
    task_dir.open('w').write('hello')
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = Trestle().run()
    assert rc > 0


@pytest.mark.parametrize(
    'sample_project, exclusions, acceptable',
    [
        (pathlib.Path('author/headers/global/good'), '', True),
        (pathlib.Path('author/headers/global/pass_with_exceptions'), '', False),
        (pathlib.Path('author/headers/global/pass_with_exceptions'), 'bad_sample', True),
        (pathlib.Path('author/headers/global/pass_with_exceptions'), 'sample_indexable_1', False),
        (pathlib.Path('author/headers/global/pass_with_nested_exceptions'), 'nested/bad', True),
        (pathlib.Path('author/headers/global/pass_with_nested_exceptions'), 'nested/good', False),
        (pathlib.Path('author/headers/global/pass_with_nested_exceptions'), 'nested', True)
    ]
)
def test_exclude_global(
    sample_project: pathlib.Path,
    exclusions: str,
    acceptable: bool,
    testdata_dir: pathlib.Path,
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour of --all flag for trestle author headers."""
    command_string = f'trestle author headers validate -r --global --trestle-root={tmp_path}'
    if not exclusions == '':
        exclusions_arr = exclusions.split(',')
        for exclusion in exclusions_arr:
            command_string += f' --exclude={exclusion.strip()}'
    # delete temp dir before copy
    # Required to be
    shutil.rmtree(tmp_path)
    shutil.copytree(str(testdata_dir / sample_project), str(tmp_path))

    template_path_original = testdata_dir / 'author' / 'headers' / 'good_templates'
    template_path_parent = tmp_path / '.trestle' / 'author'
    template_path_parent.mkdir(parents=True)
    template_path_target = template_path_parent / '__global__'
    shutil.copytree(str(template_path_original), str(template_path_target))

    monkeypatch.setattr(sys, 'argv', command_string.split())
    rc = Trestle().run()  # noqa: E800
    assert (rc == 0) == acceptable  # noqa: E800


@pytest.mark.parametrize(
    'task_name, template_path, content_path, global_, setup_rc, template_validate_rc, validate_rc',
    [
        ('catalogs', pathlib.Path('irrelevant'), pathlib.Path('irrelevant_2'), False, 1, 1, 1),
        (
            'test_tasks_1',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/meets_templates'),
            False,
            0,
            0,
            0
        ),
        (
            'test_tasks_2',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            0,
            0,
            0
        ),
        (
            'test_tasks_3',
            pathlib.Path('author/headers/bad_templates_drawio'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            0,
            1,
            1
        ),
        (
            'test_tasks_4',
            pathlib.Path('author/headers/bad_templates_wrong_names'),
            pathlib.Path('author/headers/recursive_pass'),
            False,
            0,
            1,
            1
        ),
        (
            'test_tasks_5',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_drawio'),
            False,
            0,
            0,
            3
        ),
        (
            'test_tasks_6',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_drawio_bad_file'),
            False,
            0,
            0,
            1
        ),
        (
            'test_tasks_7',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_md'),
            False,
            0,
            0,
            3
        ),
        (
            'test_tasks_8',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/fails_validation_md_bad_file'),
            False,
            0,
            0,
            3
        ),
        (
            'test_tasks_9',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/passes_with_extraneous_files'),
            False,
            0,
            0,
            0
        ),
        (
            'test_tasks_10',
            pathlib.Path('author/headers/bad_templates_md'),
            pathlib.Path('author/headers/passes_with_extraneous_files'),
            False,
            0,
            1,
            1
        ),
        (
            'test_tasks_11',
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            True,
            0,
            0,
            0
        ),
        (
            None,
            pathlib.Path('author/headers/good_templates'),
            pathlib.Path('author/headers/recursive_pass'),
            True,
            0,
            0,
            0
        )
    ]
)
def test_e2e_backward_compatibility(
    task_name: str,
    template_path: pathlib.Path,
    content_path: pathlib.Path,
    global_: bool,
    setup_rc: int,
    template_validate_rc: int,
    validate_rc: int,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Run an E2E workflow with existing files."""
    command_string_setup = 'trestle author headers setup'
    command_string_create_sample = 'trestle author headers create-sample'
    command_string_validate_template = 'trestle author headers template-validate'
    command_string_validate_content = 'trestle author headers validate -r'
    template_target_loc: pathlib.Path
    if task_name:
        tn_string = f' -tn {task_name}'
        command_string_setup += tn_string
        command_string_create_sample += tn_string
        command_string_validate_template += tn_string
        command_string_validate_content += tn_string
        template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / START_TEMPLATE_VERSION
        old_template_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name
    else:
        task_name = 'placeholder'
    if global_:
        global_str = ' --global'
        command_string_setup += global_str
        command_string_create_sample += global_str
        command_string_validate_template += global_str
        command_string_validate_content += global_str
        template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / '__global__' / START_TEMPLATE_VERSION
        old_template_loc = tmp_trestle_dir / '.trestle' / 'author' / '__global__'
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}'

    # Copy in template using old path
    template_loc = testdata_dir / template_path
    if template_loc.is_dir():
        shutil.copytree(str(template_loc), str(old_template_loc))
    else:
        if not old_template_loc.exists():
            old_template_loc.mkdir(parents=True)
        if template_loc.is_file():
            shutil.copy2(str(template_loc), str(old_template_loc))

    assert old_template_loc.exists()
    assert not template_target_loc.exists()

    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = Trestle().run()
    assert rc == setup_rc
    if setup_rc > 0:
        return

    all_files_wo_version = list(
        filter(lambda p: p.is_file() and not file_utils.is_hidden(p), (old_template_loc.iterdir()))
    )
    all_files_w_version = list(
        filter(lambda p: p.is_file() and not file_utils.is_hidden(p), (template_target_loc.iterdir()))
    )

    if template_loc.is_dir():
        all_old_files = list(filter(lambda p: p.is_file() and not file_utils.is_hidden(p), (template_loc.iterdir())))
        all_old_files = [el.parts[-1] for el in all_old_files]
    else:
        all_old_files = [template_loc.parts[-1]]

    assert len(all_files_wo_version) == 0
    assert template_target_loc.exists()
    assert all(el in [el.parts[-1] for el in all_files_w_version] for el in all_old_files)

    shutil.rmtree(str(template_target_loc))
    shutil.copytree(str(testdata_dir / template_path), str(template_target_loc))

    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    rc = Trestle().run()
    assert rc == template_validate_rc
    if template_validate_rc > 0:
        return

    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    rc = Trestle().run()
    assert rc == 0

    shutil.copytree(str(testdata_dir / content_path), str(test_content_loc))
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == validate_rc


def test_ignore_flag(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that ignored files are not validated. Validation will fail if attempted."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/headers/good_templates'
    test_instances_folder = testdata_dir / 'author/headers/ignored_files'
    task_instance_folder = tmp_trestle_dir / 'test_task'

    shutil.copytree(test_template_folder, task_template_folder)

    # copy all files
    shutil.copytree(test_instances_folder, task_instance_folder)

    command_string_validate_content = 'trestle author headers validate -tn test_task -ig ^_.*'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == 0

    # test recursive validation
    command_string_validate_content = 'trestle author headers validate -tn test_task -ig ^_.* -r'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == 0


def test_instance_no_header(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating instance with no header."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/headers/good_templates'
    test_instances_folder = testdata_dir / 'author/headers/empty_headers'
    task_instance_folder = tmp_trestle_dir / 'test_task'

    shutil.copytree(test_template_folder, task_template_folder)

    # copy all files
    shutil.copytree(test_instances_folder, task_instance_folder)

    command_string_validate_content = 'trestle author headers validate -tn test_task'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == 3

    command_string_validate_content = 'trestle author headers validate -tn test_task -tv 0.0.1'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == 3

    command_string_validate_content = 'trestle author headers validate -tv 0.0.1'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = Trestle().run()
    assert rc == 1



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the profile author module."""

import argparse
import pathlib
import shutil
import sys
from typing import Dict, Optional, Tuple

from _pytest.monkeypatch import MonkeyPatch

import pytest

from ruamel.yaml import YAML

from tests import test_utils

import trestle.oscal.catalog as cat
import trestle.oscal.profile as prof
from trestle.cli import Trestle
from trestle.common.err import TrestleError
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog_interface import CatalogInterface
from trestle.core.commands.author.profile import ProfileAssemble, ProfileGenerate, sections_to_dict
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver

# test dicts are of form {'name_exp': [(name, exp_str)...], 'text': prose}
# the text is appended to the end of the file
# then the assembled control is searched for exp_str in the prose of the named parts

markdown_name = 'my_md'
prof_name = 'my_prof'
md_name = 'my_md'
assembled_prof_name = 'my_assembled_prof'

my_guidance_text = """

## Control My Guidance

This is My Guidance.
"""

# just add a new addition
my_guidance_dict = {'name_exp': [('my_guidance', 'This is My Guidance.')], 'text': my_guidance_text}

multi_guidance_text = my_guidance_text = """

## Control A Guidance

This is A Guidance.

## Control B Guidance

This is B Guidance.
"""
# add two additions
multi_guidance_dict = {
    'name_exp': [('a_guidance', 'This is A Guidance.'), ('b_guidance', 'This is B Guidance.')],
    'text': multi_guidance_text
}

all_sections_str = (
    'ImplGuidance:Implementation Guidance,ExpectedEvidence:Expected Evidence,my_guidance:My Guidance,'
    'a_guidance:A Guidance,b_guidance:B Guidance,NeededExtra:Needed Extra'
)

all_sections_dict = {
    'ImplGuidance': 'Implementation Guidance',
    'ExpectedEvidence': 'Expected Evidence',
    'my_guidance': 'My Guidance',
    'a_guidance': 'A Guidance',
    'b_guidance': 'B Guidance',
    'NeededExtra': 'Needed Extra'
}


def edit_files(control_path: pathlib.Path, set_parameters: bool, guid_dict: Dict[str, str]) -> None:
    """Edit the files to show assemble worked."""
    assert control_path.exists()
    assert test_utils.insert_text_in_file(control_path, None, guid_dict['text'])
    if set_parameters:
        assert test_utils.delete_line_in_file(control_path, 'label:')
        assert test_utils.insert_text_in_file(control_path, 'ac-1_prm_1:', '    label: label from edit\n')
        # delete profile values for 4, then replace value for 3 with new value
        assert test_utils.insert_text_in_file(control_path, 'officer', '    profile-values: new value\n')
        assert test_utils.delete_line_in_file(control_path, 'weekly')


def setup_profile_generate(trestle_root: pathlib.Path) -> Tuple[pathlib.Path, pathlib.Path, pathlib.Path, pathlib.Path]:
    """Set up files for profile generate."""
    nist_catalog_path = test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME
    trestle_cat_dir = trestle_root / 'catalogs/nist_cat'
    trestle_cat_dir.mkdir(exist_ok=True, parents=True)
    shutil.copy(nist_catalog_path, trestle_cat_dir / 'catalog.json')
    profile_dir = trestle_root / f'profiles/{prof_name}'
    profile_dir.mkdir(parents=True, exist_ok=True)
    # simple test profile sets values for ac-1 params 1-6 but not param_7
    simple_prof_path = test_utils.JSON_TEST_DATA_PATH / 'simple_test_profile.json'
    profile_path = profile_dir / 'profile.json'
    shutil.copy(simple_prof_path, profile_path)
    markdown_path = trestle_root / md_name
    ac1_path = markdown_path / 'ac/ac-1.md'
    assembled_prof_dir = trestle_root / f'profiles/{assembled_prof_name}'
    return ac1_path, assembled_prof_dir, profile_path, markdown_path


@pytest.mark.parametrize('add_header', [True, False])
@pytest.mark.parametrize('guid_dict', [my_guidance_dict, multi_guidance_dict])
@pytest.mark.parametrize('use_cli', [True, False])
@pytest.mark.parametrize('dir_exists', [True, False])
@pytest.mark.parametrize('set_parameters', [True, False])
def test_profile_generate_assemble(
    add_header: bool,
    guid_dict: Dict,
    use_cli: bool,
    dir_exists: bool,
    set_parameters: bool,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Test the profile markdown generator."""
    ac1_path, assembled_prof_dir, profile_path, markdown_path = setup_profile_generate(tmp_trestle_dir)
    yaml_header_path = test_utils.YAML_TEST_DATA_PATH / 'good_simple.yaml'

    # convert resolved profile catalog to markdown then assemble it after adding an item to a control
    if use_cli:
        test_args = f'trestle author profile-generate -n {prof_name} -o {md_name} -rs NeededExtra'.split()
        if add_header:
            test_args.extend(['-y', str(yaml_header_path)])
        test_args.extend(['-s', all_sections_str])
        monkeypatch.setattr(sys, 'argv', test_args)
        assert Trestle().run() == 0

        edit_files(ac1_path, set_parameters, guid_dict)

        test_args = f'trestle author profile-assemble -n {prof_name} -m {md_name} -o {assembled_prof_name}'.split()
        if set_parameters:
            test_args.append('-sp')
        if dir_exists:
            assembled_prof_dir.mkdir()
        monkeypatch.setattr(sys, 'argv', test_args)
        assert Trestle().run() == 0
    else:
        profile_generate = ProfileGenerate()
        yaml_header = {}
        if add_header:
            yaml = YAML()
            yaml_header = yaml.load(yaml_header_path.open('r'))
        sections_dict = sections_to_dict(all_sections_str)
        profile_generate.generate_markdown(
            tmp_trestle_dir, profile_path, markdown_path, yaml_header, False, sections_dict, 'NeededExtra'
        )

        edit_files(ac1_path, set_parameters, guid_dict)

        if dir_exists:
            assembled_prof_dir.mkdir()
        assert ProfileAssemble.assemble_profile(
            tmp_trestle_dir, prof_name, md_name, assembled_prof_name, set_parameters, False, None, None, None
        ) == 0

    # check the assembled profile is as expected
    profile: prof.Profile
    profile, _ = ModelUtils.load_top_level_model(tmp_trestle_dir, assembled_prof_name,
                                                 prof.Profile, FileContentType.JSON)
    assert ModelUtils.model_age(profile) < test_utils.NEW_MODEL_AGE_SECONDS
    # get the set_params in the assembled profile
    set_params = profile.modify.set_parameters
    if set_parameters:
        assert set_params[0].param_id == 'ac-1_prm_1'
        assert set_params[0].values[0].__root__ == 'all personnel'
        assert set_params[1].param_id == 'ac-1_prm_2'
        assert set_params[1].values[0].__root__ == 'Organization-level'
        assert set_params[1].values[1].__root__ == 'System-level'
        assert set_params[2].param_id == 'ac-1_prm_3'
        assert set_params[2].values[0].__root__ == 'new value'
    else:
        assert len(set_params) == 15

    # now create the resolved profile catalog from the assembled json profile and confirm the addition is there

    catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, assembled_prof_dir / 'profile.json')
    catalog_interface = CatalogInterface(catalog)
    # confirm presence of all expected strings in the control named parts
    for name, exp_str in guid_dict['name_exp']:
        prose = catalog_interface.get_control_part_prose('ac-1', name)
        assert prose.find(exp_str) >= 0


@pytest.mark.parametrize(
    'required_sections, success', [(None, True), ('a_guidance,b_guidance', True), ('a_guidance,c_guidance', False)]
)
@pytest.mark.parametrize('ohv', [True, False])
def test_profile_ohv(required_sections: Optional[str], success: bool, ohv: bool, tmp_trestle_dir: pathlib.Path) -> None:
    """Test profile generate assemble with overwrite-header-values."""
    ac1_path, assembled_prof_dir, profile_path, markdown_path = setup_profile_generate(tmp_trestle_dir)
    yaml_header_path = test_utils.YAML_TEST_DATA_PATH / 'good_simple.yaml'
    new_version = '1.2.3'

    # convert resolved profile catalog to markdown then assemble it after adding an item to a control
    # if set_parameters is true, the yaml header will contain all the parameters
    profile_generate = ProfileGenerate()
    yaml = YAML()
    yaml_header = yaml.load(yaml_header_path.open('r'))
    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, yaml_header, ohv, None, None)

    edit_files(ac1_path, True, multi_guidance_dict)
    markdown_path = tmp_trestle_dir / md_name
    # change guidance in the other two controls but don't change header
    ac2_path = markdown_path / 'ac/ac-2.md'
    ac21_path = markdown_path / 'ac/ac-2.1.md'
    edit_files(ac2_path, False, multi_guidance_dict)
    edit_files(ac21_path, False, multi_guidance_dict)

    if success:
        # always doing set_params True
        assert ProfileAssemble.assemble_profile(
            tmp_trestle_dir, prof_name, md_name, assembled_prof_name, True, False, new_version, required_sections, None
        ) == 0

        # check the assembled profile is as expected
        profile: prof.Profile
        profile, _ = ModelUtils.load_top_level_model(
            tmp_trestle_dir, assembled_prof_name,
            prof.Profile,
            FileContentType.JSON
        )
        set_params = profile.modify.set_parameters

        assert len(set_params) == 14
        assert set_params[0].values[0].__root__ == 'all personnel'
        # the label is present in the header so it ends up in the set_parameter
        assert set_params[0].label == 'label from edit'
        assert set_params[1].param_id == 'ac-1_prm_2'
        assert set_params[1].values[0].__root__ == 'Organization-level'
        assert set_params[1].values[1].__root__ == 'System-level'
        assert set_params[2].values[0].__root__ == 'new value'
        assert profile.metadata.version.__root__ == new_version
        if ohv:
            assert set_params[3].values[0].__root__ == 'no meetings'
            assert set_params[3].label == 'meetings cancelled'
        else:
            assert set_params[3].values[0].__root__ == 'all meetings'
            assert set_params[3].label is None

        catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, assembled_prof_dir / 'profile.json')
        catalog_interface = CatalogInterface(catalog)
        # confirm presence of all expected strings in the control named parts
        for name, exp_str in multi_guidance_dict['name_exp']:
            prose = catalog_interface.get_control_part_prose('ac-1', name)
            assert prose.find(exp_str) >= 0
    else:
        with pytest.raises(TrestleError):
            ProfileAssemble.assemble_profile(
                tmp_trestle_dir,
                prof_name,
                md_name,
                assembled_prof_name,
                True,
                False,
                new_version,
                required_sections,
                None
            )


def test_profile_failures(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test failure modes of profile generate and assemble."""
    # no trestle root specified direct command
    test_args = argparse.Namespace(name='my_prof', output='new_prof', verbose=0, set_parameters=False)
    profile_generate = ProfileGenerate()
    assert profile_generate._run(test_args) == 1

    # no trestle root specified
    profile_assemble = ProfileAssemble()
    assert profile_assemble._run(test_args) == 1

    # bad yaml
    bad_yaml_path = str(test_utils.YAML_TEST_DATA_PATH / 'bad_simple.yaml')
    trestle_root = tmp_trestle_dir
    test_args = argparse.Namespace(
        trestle_root=trestle_root,
        name='my_prof',
        output='new_prof',
        yaml_header=bad_yaml_path,
        verbose=0,
        set_parameters=False
    )
    profile_generate = ProfileGenerate()
    assert profile_generate._run(test_args) == 1

    # profile not available for load
    test_args = 'trestle author profile-generate -n my_prof -o my_md -v'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1

    # setup for generate and assemble
    profile_path = ModelUtils.path_for_top_level_model(tmp_trestle_dir, 'my_prof', prof.Profile, FileContentType.JSON)
    profile_path.parent.mkdir()
    shutil.copyfile(test_utils.JSON_TEST_DATA_PATH / 'simple_test_profile.json', profile_path)
    cat_path = ModelUtils.path_for_top_level_model(tmp_trestle_dir, 'nist_cat', cat.Catalog, FileContentType.JSON)
    cat_path.parent.mkdir()
    shutil.copyfile(test_utils.JSON_NIST_DATA_PATH / test_utils.JSON_NIST_CATALOG_NAME, cat_path)

    # generate markdown with required section but don't give it any prose
    test_args = argparse.Namespace(
        trestle_root=trestle_root,
        name='my_prof',
        output='md_prof',
        verbose=0,
        set_parameters=False,
        overwrite_header_values=False,
        yaml_header=None,
        sections='NeededExtra:Needed Extra,ImplGuidance:Implementation Guidance,ExpectedEvidence:Expected Evidence',
        required_sections='NeededExtra'
    )
    profile_generate = ProfileGenerate()
    assert profile_generate._run(test_args) == 0

    # set arguments for assemble
    test_args = argparse.Namespace(
        trestle_root=trestle_root,
        name=None,
        markdown='md_prof',
        output='my_prof',
        verbose=0,
        set_parameters=False,
        yaml_header=None,
        required_sections='NeededExtra',
        regenerate=False,
        version=None,
        allowed_sections=None
    )
    # fail since required section not filled in
    profile_assemble = ProfileAssemble()
    assert profile_assemble._run(test_args) == 1

    # succeed if not specifying required section
    test_args.required_sections = None
    assert profile_assemble._run(test_args) == 0

    # fail if allowed sections doesn't include all needed
    test_args.allowed_sections = 'NeededExtra,DummySection'
    assert profile_assemble._run(test_args) == 1

    # succed if allowed sections has all
    test_args.allowed_sections = 'ExpectedEvidence,ImplGuidance,NeededExtra'
    assert profile_assemble._run(test_args) == 0

    # disallowed output name
    test_args = 'trestle author profile-generate -n my_prof -o profiles -v'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1


def test_profile_overwrite(tmp_trestle_dir: pathlib.Path) -> None:
    """Test blocking overwrite if no change to assembled profile relative to one it would overwrite."""
    _, _, profile_path, markdown_path = setup_profile_generate(tmp_trestle_dir)

    # generate the markdown and assemble
    profile_generate = ProfileGenerate()
    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, {}, False, all_sections_dict, None)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, prof_name, True, False, None, None, None
    ) == 0

    # note the file timestamp, then regenerate and assemble again
    orig_time = profile_path.stat().st_mtime

    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, {}, False, all_sections_dict, None)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, prof_name, True, False, None, None, None
    ) == 0

    # the timestamp should be the same, indicating the file was not written
    new_time = profile_path.stat().st_mtime

    assert new_time == orig_time

    # now generate again but with different section title, causing change in generated profile markdown
    new_sections = {'ImplGuidance': 'Different Title'}
    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, {}, False, new_sections, None)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, prof_name, True, False, None, None, None
    ) == 0

    # the timestamp should now be different
    new_time = profile_path.stat().st_mtime

    assert new_time != orig_time



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Import Command."""
import argparse
import logging
import pathlib

import trestle.core.commands.validate as validatecmd
from trestle.common import const, file_utils, log
from trestle.common.err import TrestleError, TrestleRootError, handle_generic_command_exception
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.command_docs import CommandPlusDocs
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.models.actions import CreatePathAction, WriteFileAction
from trestle.core.models.elements import Element
from trestle.core.models.file_content_type import FileContentType
from trestle.core.models.plans import Plan
from trestle.core.remote import cache

logger = logging.getLogger(__name__)


class ImportCmd(CommandPlusDocs):
    """Import an existing full OSCAL model into the trestle project."""

    name = 'import'

    def _init_arguments(self) -> None:
        logger.debug('Init arguments')
        self.add_argument(
            '-f', '--file', help='OSCAL file to import - either file path or url.', type=str, required=True
        )
        self.add_argument('-o', '--output', help='Name of output element.', type=str, required=True)
        self.add_argument('-r', '--regenerate', action='store_true', help=const.HELP_REGENERATE)

    def _run(self, args: argparse.Namespace) -> int:
        """Top level import run command."""
        try:
            log.set_log_level_from_args(args)
            trestle_root = args.trestle_root
            if not file_utils.is_valid_project_root(trestle_root):
                raise TrestleRootError(f'Attempt to import from non-valid trestle project root {trestle_root}')

            input_uri = args.file
            if cache.FetcherFactory.in_trestle_directory(trestle_root, input_uri):
                raise TrestleError(
                    f'Imported file {input_uri} cannot be from current trestle project. Use duplicate instead.'
                )

            content_type = FileContentType.to_content_type('.' + input_uri.split('.')[-1])

            fetcher = cache.FetcherFactory.get_fetcher(trestle_root, str(input_uri))

            model_read, parent_alias = fetcher.get_oscal(True)

            plural_path = ModelUtils.model_type_to_model_dir(parent_alias)

            output_name = args.output

            desired_model_dir = trestle_root / plural_path
            desired_model_path: pathlib.Path = desired_model_dir / output_name / parent_alias
            desired_model_path = desired_model_path.with_suffix(FileContentType.to_file_extension(content_type)
                                                                ).resolve()

            if desired_model_path.exists():
                raise TrestleError(
                    f'Cannot import because file to be imported here: {desired_model_path} already exists.'
                )

            if args.regenerate:
                logger.debug(f'regenerating uuids in imported file {input_uri}')
                model_read, lut, nchanged = ModelUtils.regenerate_uuids(model_read)
                logger.debug(f'uuid lut has {len(lut.items())} entries and {nchanged} refs were updated')

            top_element = Element(model_read)
            create_action = CreatePathAction(desired_model_path, True)
            write_action = WriteFileAction(desired_model_path, top_element, content_type)

            # create a plan to create the directory and write the imported file.
            import_plan = Plan()
            import_plan.add_action(create_action)
            import_plan.add_action(write_action)

            import_plan.execute()

            args = argparse.Namespace(
                file=desired_model_path, verbose=args.verbose, trestle_root=args.trestle_root, type=None, all=None
            )
            rollback = False
            try:
                rc = validatecmd.ValidateCmd()._run(args)
                if rc > 0:
                    logger.warning(f'Validation of imported file {desired_model_path} did not pass')
                    rollback = True
            except TrestleError as err:
                logger.warning(f'Import of {str(input_uri)} failed with validation error: {err}')
                rollback = True

            if rollback:
                logger.debug(f'Rolling back import of {str(input_uri)} to {desired_model_path}')
                try:
                    import_plan.rollback()
                except TrestleError as err:
                    raise TrestleError(
                        f'Import failed in plan rollback: {err}. Manually remove {desired_model_path} to recover.'
                    )
                logger.debug(f'Successful rollback of import to {desired_model_path}')
                return CmdReturnCodes.COMMAND_ERROR.value

            return CmdReturnCodes.SUCCESS.value

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while importing OSCAL file')



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model parsing for use when models themselves must be inferred and are not known.

Under most use cases trestle.core.base_model.OscalBaseModel provides functionality for loading Oscal models from files.
However, under some circumstances the model internals are unknown.  Use of this module should be avoided unless
the BaseModel functionality is inadequate.
"""

import importlib
import logging
from typing import Any, Dict

from trestle.common import const
from trestle.common.err import TrestleError
from trestle.common.str_utils import AliasMode, alias_to_classname
from trestle.core.base_model import OscalBaseModel

logger = logging.getLogger(__name__)


def parse_dict(data: Dict[str, Any], model_name: str) -> OscalBaseModel:
    """Load a model from the data dict.

    This functionality is provided for situations when the OSCAL data type is not known ahead of time. Here the model
    has been loaded into memory using json loads or similar and passed as a dict.

    Args:
        data: Oscal data loaded into memory as a dictionary with the `root key` removed.
        model_name: should be of the form 'module.class' from trestle.oscal.* modules

    Returns:
        The oscal model of the desired model.
    """
    if data is None:
        raise TrestleError('data name is required')

    if model_name is None:
        raise TrestleError('model_name is required')

    parts = model_name.split('.')
    class_name = parts.pop()
    module_name = '.'.join(parts)

    logger.debug(f'Loading class "{class_name}" from "{module_name}"')
    module = importlib.import_module(module_name)
    mclass: OscalBaseModel = getattr(module, class_name)
    if mclass is None:
        raise TrestleError(f'class "{class_name}" could not be found in "{module_name}"')

    instance = mclass.parse_obj(data)
    return instance


def root_key(data: Dict[str, Any]) -> str:
    """Find root model name in the data."""
    if len(data.items()) == 1:
        return next(iter(data))

    raise TrestleError('data does not contain a root key')


def to_full_model_name(root_key: str) -> str:
    """
    Find model name from the root_key in the file.

    Args:
        root_key: root key such as 'system-security-plan' from a top level OSCAL model.
    """
    if root_key not in const.MODEL_TYPE_LIST:
        raise TrestleError(f'{root_key} is not a top level model name.')

    module = const.MODEL_TYPE_TO_MODEL_MODULE[root_key]
    class_name = alias_to_classname(root_key, AliasMode.JSON)
    return f'{module}.{class_name}'



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for validator helper functionality that was moved to model_utils."""

import pathlib
from uuid import uuid4

import tests.test_utils as test_utils

import trestle.oscal.catalog as catalog
import trestle.oscal.ssp as ssp
from trestle.common.model_utils import ModelUtils

ssp_path = pathlib.Path('nist-content/src/examples/ssp/json/ssp-example.json')
catalog_path = test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME


def test_regenerate_uuids_ssp() -> None:
    """Test regeneration of uuids with updated refs in ssp."""
    orig_ssp = ssp.SystemSecurityPlan.oscal_read(ssp_path)
    new_ssp, uuid_lut, n_refs_updated = ModelUtils.regenerate_uuids(orig_ssp)
    assert len(uuid_lut.items()) == 36
    assert n_refs_updated == 23


def test_regenerate_uuids_catalog() -> None:
    """Test regeneration of uuids with updated refs in catalog."""
    orig_cat = catalog.Catalog.oscal_read(catalog_path)
    new_cat, uuid_lut, n_refs_updated = ModelUtils.regenerate_uuids(orig_cat)
    assert len(uuid_lut.items()) == 2
    assert n_refs_updated == 2


def test_validations_on_dict() -> None:
    """Test regen of uuid in dict."""
    my_uuid1 = str(uuid4())
    my_uuid2 = str(uuid4())
    my_dict = {'uuid': my_uuid1, 'ref': my_uuid1, 'my_inner_dict': {'uuid': my_uuid2, 'ref': my_uuid2}}
    new_dict, lut = ModelUtils._regenerate_uuids_in_place(my_dict, {})
    assert my_dict['uuid'] != new_dict['uuid']
    assert my_dict['my_inner_dict']['uuid'] != new_dict['my_inner_dict']['uuid']
    assert len(lut) == 2

    fixed_dict, count = ModelUtils._update_new_uuid_refs(new_dict, lut)
    assert fixed_dict['uuid'] == fixed_dict['ref']
    assert fixed_dict['my_inner_dict']['uuid'] == fixed_dict['my_inner_dict']['ref']
    assert count == 2


def test_find_by_name() -> None:
    """Test finding by name in dict."""
    my_dict = {'a': 'foo', 'b': 'bar', 'c': {'x': 1, 'a': 7, 'y': 'hello'}}
    result = ModelUtils.find_values_by_name(my_dict, 'a')
    assert result == ['foo', 7]



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
#
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#                                                                        #
#        This file is automatically generated by gen_oscal.py            #
#                                                                        #
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#
#
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import AnyUrl, EmailStr, Extra, Field, conint, constr, validator

from trestle.core.base_model import OscalBaseModel
from trestle.oscal import OSCAL_VERSION_REGEX, OSCAL_VERSION
import trestle.oscal.common as common


class Control(OscalBaseModel):
    """
    A structured information object representing a security or privacy control. Each security or privacy control within the Catalog is defined by a distinct control instance.
    """

    class Config:
        extra = Extra.forbid

    id: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        description=
        'A human-oriented, locally unique identifier with instance scope that can be used to reference this control elsewhere in this and other OSCAL instances (e.g., profiles). This id should be assigned per-subject, which means it should be consistently used to identify the same control across revisions of the document.',
        title='Control Identifier',
    )
    class_: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        alias='class',
        description='A textual label that provides a sub-type or characterization of the control.',
        title='Control Class',
    )
    title: str = Field(
        ...,
        description='A name given to the control, which may be used by a tool for display and navigation.',
        title='Control Title',
    )
    params: Optional[List[common.Parameter]] = Field(None)
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    parts: Optional[List[common.Part]] = Field(None)
    controls: Optional[List[Control]] = None


class Group(OscalBaseModel):
    """
    A group of controls, or of groups of controls.
    """

    class Config:
        extra = Extra.forbid

    id: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        description=
        'A human-oriented, locally unique identifier with cross-instance scope that can be used to reference this defined group elsewhere in in this and other OSCAL instances (e.g., profiles). This id should be assigned per-subject, which means it should be consistently used to identify the same group across revisions of the document.',
        title='Group Identifier',
    )
    class_: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        alias='class',
        description='A textual label that provides a sub-type or characterization of the group.',
        title='Group Class',
    )
    title: str = Field(
        ...,
        description='A name given to the group, which may be used by a tool for display and navigation.',
        title='Group Title',
    )
    params: Optional[List[common.Parameter]] = Field(None)
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    parts: Optional[List[common.Part]] = Field(None)
    groups: Optional[List[Group]] = None
    controls: Optional[List[Control]] = Field(None)


class Catalog(OscalBaseModel):
    """
    A collection of controls.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A globally unique identifier with cross-instance scope for this catalog instance. This UUID should be changed when this document is revised.',
        title='Catalog Universally Unique Identifier',
    )
    metadata: common.Metadata
    params: Optional[List[common.Parameter]] = Field(None)
    controls: Optional[List[Control]] = Field(None)
    groups: Optional[List[Group]] = Field(None)
    back_matter: Optional[common.BackMatter] = Field(None, alias='back-matter')


class Model(OscalBaseModel):
    catalog: Catalog


Control.update_forward_refs()
Group.update_forward_refs()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for json and yaml manipulations of oscal files."""

import datetime
import pathlib

import pytest

from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError

import trestle.common.const as const
import trestle.oscal.component as component

yaml_path = pathlib.Path('tests/data/yaml/')
json_path = pathlib.Path('tests/data/json/')
encoding = const.FILE_ENCODING


def test_yaml_load() -> None:
    """Test yaml load."""
    # happy path
    read_file = (yaml_path / 'good_simple.yaml').open('r', encoding=encoding)
    yaml = YAML(typ='safe')
    obj = yaml.load(read_file)
    assert obj is not None

    # unhappy path
    with pytest.raises(ParserError):
        read_file = (yaml_path / 'bad_simple.yaml').open('r', encoding=encoding)
        obj = yaml.load(read_file)


def test_yaml_dump(tmp_path: pathlib.Path) -> None:
    """Test yaml load and dump."""
    component_name = 'good_component.yaml'
    tmp_path = pathlib.Path(tmp_path)
    yaml = YAML(typ='safe')
    # happy path
    read_file = (yaml_path / component_name).open('r', encoding=encoding)
    component_obj = yaml.load(read_file)
    read_file.close()
    assert component_obj is not None

    dump_name = tmp_path / component_name
    write_file = dump_name.open('w', encoding=encoding)
    yaml.dump(component_obj, write_file)
    write_file.close()
    read_file = dump_name.open('r', encoding=encoding)
    saved_component = yaml.load(read_file)
    assert saved_component is not None

    assert saved_component == component_obj


def test_oscal_model(tmp_path: pathlib.Path) -> None:
    """Test pydantic oscal model."""
    good_component_name = 'good_component.yaml'
    tmp_path = pathlib.Path(tmp_path)

    tmp_path = pathlib.Path(tmp_path)

    # load good component
    read_file = yaml_path / good_component_name
    assert read_file.exists()
    component_obj = component.ComponentDefinition.oscal_read(read_file)
    assert component_obj is not None

    # write the oscal componet def out as yaml
    dump_name = tmp_path / good_component_name
    component_obj.oscal_write(dump_name)

    # read it back in
    component_reload = component.ComponentDefinition.oscal_read(dump_name)
    assert component_reload is not None

    # confirm same
    assert component_obj == component_reload

    # confirm it really is checking the time
    component_reload.metadata.last_modified = datetime.datetime.now()
    assert component_obj != component_reload

    # load good target with different timezone
    read_file = yaml_path / 'good_component_diff_tz.yaml'
    component_diff_tz = component.ComponentDefinition.oscal_read(read_file)
    assert component_diff_tz is not None

    # confirm same since different timezones but same utc time
    assert component_obj == component_diff_tz

    # try to load file with no timezone specified
    read_file = yaml_path / 'bad_component_no_tz.yaml'

    # confirm the load fails because it is invalid without timezone specified
    try:
        _ = component.ComponentDefinition.oscal_read(read_file)
    except Exception:
        assert True
    else:
        assert AssertionError()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script to remove references in json schema to simplify work done by datamodel-generator."""
import json
import logging
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


class FlattenSchema():
    """Flatten the schema by resolving references as much as possible."""

    _ref_str = '$ref'
    _refd = set()
    _def_list = []

    def _find_ref(self, name):
        """Find the full definition of the given reference name from the list of definitions."""
        for (key, value) in self._def_list:
            if key == name:
                return value
        raise AssertionError()

    @staticmethod
    def _get_ref_name(s):
        """Extract the name from the ref definition value."""
        pref_string = '#/definitions/'
        assert s.find(pref_string) == 0
        return s[len(pref_string):]

    def _replace_ref(self, d, ref_set):
        """Input is known to be ref: /definitions/bar so just return replacement def from def_list."""
        ref_name = FlattenSchema._get_ref_name(d[self._ref_str])
        if ref_name in ref_set:  # name is already on the stack so stop now and don't recurse
            return d, ref_set, False
        # this is a new reference so return the body of the definition and add its name as a reference
        self._refd.add(ref_name)
        ref_set.add(ref_name)
        return self._find_ref(ref_name), ref_set, True

    def _replace_refs(self, obj, ref_set):
        """Given an object recurse into it replacing any found ref: defs with what is in def_list."""
        if type(obj) == dict:
            # first check if it is a simple $ref line and replace it directly
            if len(obj.items()) == 1 and obj.get(self._ref_str, None) is not None:
                return self._replace_ref(obj, ref_set)
            new_dict = {}
            dirty = False
            for key, val in obj.items():
                if key in ref_set:
                    new_dict[key] = val
                else:
                    ref_set.add(key)
                    new_dict[key], ref_set, changed = self._replace_refs(val, ref_set)
                    if changed:
                        dirty = True
            return new_dict, ref_set, dirty
        elif type(obj) == str:
            return obj, ref_set, False
        elif type(obj) == list:
            n_list = len(obj)
            changed = False
            dirty = False
            for i in range(n_list):
                obj[i], ref_set, changed = self._replace_refs(obj[i], ref_set)
                if changed:
                    dirty = True
            return obj, ref_set, dirty
        elif type(obj) == tuple:
            new_val, ref_set, changed = self._replace_refs(obj[1], ref_set)
            return (obj[0], new_val), ref_set, changed
        if hasattr(obj, '__iter__'):
            logger.info('missed iterable type: ', type(obj))
        return obj, ref_set, False

    def _replace_schema_refs(self, schema):
        """Expand each definition with repeated iterations over the list with recursion into each."""
        self._def_list = list(schema['definitions'].items())
        n_defs = len(self._def_list)
        fixed = [False] * n_defs
        dirty = True
        while dirty:
            dirty = False
            for i in range(n_defs):
                changed = False
                if not fixed[i]:
                    self._def_list[i], _, changed = self._replace_refs(self._def_list[i], {self._def_list[i][0]})
                    if not changed:  # mark it fixed and don't revisit
                        fixed[i] = True
            if changed:
                dirty = True
        new_dict = {}
        for tup in self._def_list:
            new_dict[tup[0]] = tup[1]
        schema['definitions'] = new_dict
        return schema

    def replace_refs(self, schema_file_name, output_file_name):
        """Replace all refs in schema file name and output with refs removed except for cyclic ones."""
        with open(schema_file_name, encoding='utf8') as f:
            schema = json.load(f)

        flat_schema = self._replace_schema_refs(schema)

        tmp_name = output_file_name + '.tmp'
        with open(tmp_name, 'w', encoding='utf8') as f:
            json.dump(flat_schema, f, indent=4)

        needed_refs = set()
        ref_str = '"$ref": "#/definitions/'
        with open(tmp_name, 'r', encoding='utf8') as fin:
            for line in fin.readlines():
                nref = line.find(ref_str)
                if nref >= 0:
                    new_ref = line[(nref + len(ref_str)):].strip()
                    needed_refs.add(new_ref)

        out_lines = []
        with open(tmp_name, 'r', encoding='utf8') as fin:
            for line in fin.readlines():
                skipit = False
                if line.find('"$id": "#/definitions/') >= 0:
                    skipit = True
                    for cname in needed_refs:
                        # skip any definitions
                        if line.find(f'"$id": "#/definitions/{cname}') >= 0:
                            skipit = False
                            break
                if not skipit:
                    out_lines.append(line)

        with open(output_file_name, 'w', encoding='utf8') as fout:
            fout.writelines(out_lines)

        os.remove(tmp_name)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle href command."""
import os
import pathlib
import sys

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils

from trestle.cli import Trestle
from trestle.common.model_utils import ModelUtils
from trestle.core.models.file_content_type import FileContentType
from trestle.oscal import profile


def test_href_cmd(
    tmp_path: pathlib.Path, keep_cwd: pathlib.Path, simplified_nist_profile: profile.Profile, monkeypatch: MonkeyPatch
) -> None:
    """Test basic cmd invocation of href."""
    # prepare trestle project dir with the file
    models_path, profile_path = test_utils.prepare_trestle_project_dir(
        tmp_path,
        FileContentType.JSON,
        simplified_nist_profile,
        test_utils.PROFILES_DIR)

    os.chdir(models_path)

    # just list the hrefs
    cmd_string = 'trestle href -n my_test_model'
    monkeypatch.setattr(sys, 'argv', cmd_string.split())
    rc = Trestle().run()
    assert rc == 0

    orig_href = simplified_nist_profile.imports[0].href

    new_href = 'trestle://catalogs/my_catalog/catalog.json'

    cmd_string = f'trestle href -n my_test_model -hr {new_href}'
    monkeypatch.setattr(sys, 'argv', cmd_string.split())
    rc = Trestle().run()
    assert rc == 0

    # confirm new href is correct
    new_profile: profile.Profile = profile.Profile.oscal_read(profile_path)
    assert new_profile.imports[0].href == new_href

    # restore orig href to confirm models are otherwise equivalent
    # only thing different should be last-modified
    new_profile.imports[0].href = orig_href
    assert ModelUtils.models_are_equivalent(new_profile, simplified_nist_profile)


def test_href_failures(
    tmp_path: pathlib.Path, keep_cwd: pathlib.Path, simplified_nist_profile: profile.Profile, monkeypatch: MonkeyPatch
) -> None:
    """Test href failure modes."""
    # prepare trestle project dir with the file
    models_path, profile_path = test_utils.prepare_trestle_project_dir(
        tmp_path,
        FileContentType.JSON,
        simplified_nist_profile,
        test_utils.PROFILES_DIR)

    cmd_string = 'trestle href -n my_test_model -hr foobar'

    # not in trestle project so fail
    monkeypatch.setattr(sys, 'argv', cmd_string.split())
    rc = Trestle().run()
    assert rc == 5

    os.chdir(models_path)

    cmd_string = 'trestle href -n my_test_model -hr foobar -i 2'

    # add extra import to the profile and ask for import number 2
    simplified_nist_profile.imports.append(simplified_nist_profile.imports[0])
    simplified_nist_profile.oscal_write(profile_path)
    monkeypatch.setattr(sys, 'argv', cmd_string.split())
    rc = Trestle().run()
    assert rc == 1



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script reduces lists in the NIST 800-53 catalog to at most two elements."""

import argparse
import logging
import pathlib

import ilcli

import trestle.cli
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.oscal import catalog as oscatalog

logger = logging.getLogger(__name__)


def prune_lists(input_model: trestle.core.base_model.OscalBaseModel, list_cap: int) -> None:
    """Recurse down into dictionaries and up to the Nth element of lists, where N is the given list_cap."""
    for _, v in input_model:
        if isinstance(v, trestle.core.base_model.OscalBaseModel):
            prune_lists(v, list_cap)
        elif isinstance(v, list):
            if len(v) > list_cap:
                del v[-(len(v) - list_cap):]
            for x in range(len(v)):
                if isinstance(v[x], trestle.core.base_model.OscalBaseModel):
                    prune_lists(v[x], list_cap)


class SimplifyCatalog(ilcli.Command):
    """Reduces lists in the NIST 800-53 catalog to at most two elements."""

    def _init_arguments(self) -> None:
        self.add_argument('input', nargs=1, help='input json to read from')
        self.add_argument('output', nargs=1, help='output json to write to')
        self.add_argument('cap', nargs='?', type=int, default=None, help='max length of lists, default=2')

    def _run(self, args: argparse.Namespace) -> int:
        catalog_file = pathlib.Path(args.input[0])

        try:
            catalog = oscatalog.Catalog.oscal_read(catalog_file)
        except Exception as e:
            logger.error(f'Failure in oscal_read({catalog_file}): {e}\n')
            return CmdReturnCodes.COMMAND_ERROR.value

        if not catalog:
            logger.error('No data in input or input read failed.\n')
            return CmdReturnCodes.COMMAND_ERROR.value

        list_cap = args.cap if args.cap else 2
        try:
            prune_lists(input_model=catalog, list_cap=list_cap)
        except Exception as e:
            logger.error(f'Problem with prune_lists(): {e}\n')
            return CmdReturnCodes.COMMAND_ERROR.value

        simplified_catalog_file = pathlib.Path(args.output[0])

        try:
            catalog.oscal_write(simplified_catalog_file)
        except Exception as e:
            logger.error(f'Problem with oscal_write: {e}\n')
            return CmdReturnCodes.COMMAND_ERROR.value

        return CmdReturnCodes.SUCCESS.value


if __name__ == '__main__':
    exit(SimplifyCatalog().run())



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle remove action class."""

from typing import List

from trestle.core.models.actions import RemoveAction, UpdateAction
from trestle.core.models.elements import Element, ElementPath
from trestle.oscal import common, component


def prepare_element(sample_nist_component_def: component.ComponentDefinition):
    """Prepare a target element for remove tests."""
    element = Element(sample_nist_component_def)

    parties: List[common.Party] = []
    parties.append(
        common.Party(**{
            'uuid': 'ff47836c-877c-4007-bbf3-c9d9bd805000', 'name': 'TEST1', 'type': 'organization'
        })
    )
    parties.append(
        common.Party(**{
            'uuid': 'ee88836c-877c-4007-bbf3-c9d9bd805000', 'name': 'TEST2', 'type': 'organization'
        })
    )

    sub_element_path = ElementPath('component-definition.metadata.parties.*')
    ac = UpdateAction(parties, element, sub_element_path)
    ac.execute()

    assert element.get_at(sub_element_path) == parties

    return element


def test_remove_action(sample_nist_component_def):
    """Test remove action."""
    element = prepare_element(sample_nist_component_def)
    sub_element_path = ElementPath('component-definition.metadata.parties')
    prev_sub_element = element.get_at(sub_element_path)

    rac = RemoveAction(element, sub_element_path)

    rac.execute()

    assert element.get_at(sub_element_path) is None

    rac.rollback()

    assert element.get_at(sub_element_path) == prev_sub_element



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plan of action of a command."""
import logging
from io import UnsupportedOperation
from typing import List

from .actions import Action

logger = logging.getLogger(__name__)


class Plan:
    """Plan of action of a command."""

    def __init__(self) -> None:
        """Initialize a plan."""
        self._actions: List[Action] = []

    def _action_key(self, action: Action) -> int:
        return hash(action)

    def __str__(self) -> str:
        """Print the plan."""
        list_actions = []
        index = 1
        for action in self._actions:
            list_actions.append(f'{index}. {action}')
            index = index + 1

        list_str = '\n'.join(list_actions)
        return list_str

    def get_actions(self) -> List[Action]:
        """Get all actions."""
        return self._actions

    def add_action(self, action: Action) -> None:
        """Add a new action."""
        self._actions.append(action)

    def add_actions(self, actions: List[Action]) -> None:
        """Add actions in order."""
        self._actions.extend(actions)

    def clear_actions(self) -> None:
        """Clear all actions."""
        self._actions = []

    def execute(self) -> None:
        """Execute the actions in the plan."""
        for action in self._actions:
            try:
                action.execute()
            except Exception as e:
                logger.error(f'Failed to execute action {action} for the plan: {e}. Rolling back.')
                self.rollback()
                raise e

    def rollback(self) -> None:
        """Rollback the actions in the plan."""
        # execute in reverse order
        for action in reversed(self._actions):
            if action.has_rollback() is False:
                raise UnsupportedOperation(f'{action.get_type()} does not support rollback')
            action.rollback()

    def __eq__(self, other: object) -> bool:
        """Check that two plans are equal."""
        if not isinstance(other, Plan):
            return False

        return self.get_actions() == other.get_actions()
